@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005-2008 Douglas Gilbert.
+ * Copyright (c) 2005-2010 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -56,19 +56,20 @@
 #include "port_getopt.h"
 #endif
 
-#ifdef SDPARM_LINUX
+#ifdef SG_LIB_LINUX
 #include <errno.h>
 #include <sys/ioctl.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/utsname.h>
 
+/* Following needed for lk 2.4 series; may be dropped in future */
 #include <scsi/scsi.h>
 #include <scsi/sg.h>
-
 static int map_if_lk24(int sg_fd, const char * device_name, int rw,
                        int verbose);
-#endif  /* SDPARM_LINUX */
+
+#endif  /* SG_LIB_LINUX */
 
 #include "sdparm.h"
 #include "sg_lib.h"
@@ -76,7 +77,7 @@ static int map_if_lk24(int sg_fd, const char * device_name, int rw,
 
 #define MAX_DEV_NAMES 256
 
-static char * version_str = "1.03 20080623";
+static char * version_str = "1.05 20100413 [svn: r149]";
 
 
 static struct option long_options[] = {
@@ -97,13 +98,14 @@ static struct option long_options[] = {
     {"num-desc", no_argument, 0, 'n'},
     {"page", required_argument, 0, 'p'},
     {"quiet", no_argument, 0, 'q'},
+    {"readonly", no_argument, 0, 'r'},
     {"set", required_argument, 0, 's'},
     {"save", no_argument, 0, 'S'},
     {"transport", required_argument, 0, 't'},
     {"vendor", required_argument, 0, 'M'},
     {"verbose", no_argument, 0, 'v'},
     {"version", no_argument, 0, 'V'},
-#ifdef SDPARM_WIN32
+#ifdef SG_LIB_WIN32
     {"wscan", no_argument, 0, 'w'},
 #endif
     {0, 0, 0, 0},
@@ -118,10 +120,10 @@ usage()
           "              [--dummy] [--flexible] [--get=STR] [--help] "
           "[--hex] [--inquiry]\n"
           "              [--long] [--num-desc] [--page=PG[,SPG]] [--quiet] "
-          "[--save]\n"
-          "              [--set=STR] [--six] [--transport=TN] [--vendor=VN] "
-          "[--verbose]\n"
-          "              [--version] DEVICE [DEVICE...]\n\n"
+          "[--readonly]\n"
+          "              [--save] [--set=STR] [--six] [--transport=TN] "
+          "[--vendor=VN]\n"
+          "              [--verbose] [--version] DEVICE [DEVICE...]\n\n"
           "       sdparm --enumerate [--all] [--inquiry] [--long] "
           "[--page=PG[,SPG]]\n"
           "              [--transport=TN] [--vendor=VN]\n"
@@ -153,6 +155,10 @@ usage()
           "enumerate\n"
           "    --quiet | -q          suppress device vendor/product/"
           "revision string line\n"
+          "    --readonly | -r       force read-only open of DEVICE (def: "
+          "either\n"
+          "                          read-write or read-only). For ATA "
+          "disks\n"
           "    --save | -S           place mode changes in saved page as "
           "well\n"
           "    --set=STR | -s STR    set field value(s)\n"
@@ -163,7 +169,7 @@ usage()
           "    --vendor=VN | -M VN    vendor (manufacturer) number "
           "[or abbrev]\n"
           "    --verbose | -v        increase verbosity\n"
-#ifdef SDPARM_WIN32
+#ifdef SG_LIB_WIN32
           "    --version | -V        print version string and exit\n"
           "    --wscan | -w          windows scan for device names\n\n"
 #else
@@ -259,7 +265,7 @@ print_mp_extra(const char * extra)
     }
     printf("\t%s\n", p);
 }
-
+/* Enumerate mode page items. Reading from internal dataset. */
 static void
 enumerate_mitems(int pn, int spn, int pdt,
                  const struct sdparm_opt_coll * opts)
@@ -369,6 +375,8 @@ list_mp_settings(const struct sdparm_mode_page_settings * mps, int get)
     }
 }
 
+/* Print prefix and acronymn plus current, changeable, default and saved
+ * values for an item */
 static void
 print_mp_entry(const char * pre, int smask,
                const struct sdparm_mode_page_item *mpi,
@@ -438,6 +446,20 @@ print_mp_entry(const char * pre, int smask,
         print_mp_extra(mpi->extra);
 }
 
+static void
+print_mp_arr_entry(const char * pre, int smask,
+                   const struct sdparm_mode_page_item *mpi,
+                   void ** pc_arr, int long_out, int force_decimal)
+{
+    const unsigned char * cur_mp = (const unsigned char *)pc_arr[0];
+    const unsigned char * cha_mp = (const unsigned char *)pc_arr[1];
+    const unsigned char * def_mp = (const unsigned char *)pc_arr[2];
+    const unsigned char * sav_mp = (const unsigned char *)pc_arr[3];
+
+    print_mp_entry(pre, smask, mpi, cur_mp, cha_mp, def_mp, sav_mp,
+                   long_out, force_decimal);
+}
+
 static int
 ll_mode_sense(int fd, const struct sdparm_opt_coll * opts, int pn, int spn,
               unsigned char * resp, int mx_resp_len, int noisy, int verb)
@@ -450,6 +472,7 @@ ll_mode_sense(int fd, const struct sdparm_opt_coll * opts, int pn, int spn,
                                   spn, resp, mx_resp_len, noisy, verb);
 }
 
+/* Make some noise if the mode page response seems badly formed */
 static void
 check_mode_page(unsigned char * cur_mp, int pn, int rep_len,
                 const struct sdparm_opt_coll * opts)
@@ -492,11 +515,11 @@ print_mpage_extra_desc(void ** pc_arr, int rep_len,
 {
     const struct sdparm_mode_descriptor_t * mdp = mpp->mp_desc;
     const struct sdparm_mode_page_item * mpi;
-    unsigned char * cur_mp = (unsigned char *)pc_arr[0];
-    unsigned char * cha_mp = (unsigned char *)pc_arr[1];
-    unsigned char * def_mp = (unsigned char *)pc_arr[2];
-    unsigned char * sav_mp = (unsigned char *)pc_arr[3];
-    unsigned char * ucp;
+    const unsigned char * cur_mp = (const unsigned char *)pc_arr[0];
+    const unsigned char * cha_mp = (const unsigned char *)pc_arr[1];
+    const unsigned char * def_mp = (const unsigned char *)pc_arr[2];
+    const unsigned char * sav_mp = (const unsigned char *)pc_arr[3];
+    const unsigned char * ucp;
     struct sdparm_mode_page_item ampi;
     uint64_t u;
     int k, num, len, d_off, n, bad;
@@ -505,9 +528,12 @@ print_mpage_extra_desc(void ** pc_arr, int rep_len,
     if ((NULL == mdp) || (NULL == cur_mp) || (rep_len < 4) ||
         (mdp->num_descs_off >= rep_len))
         return;
-    u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
-                           mdp->num_descs_bytes * 8) +
-        mdp->num_descs_inc;
+    if ((mdp->num_descs_inc < 0) && (mdp->desc_len > 0))
+        u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
+                       mdp->num_descs_bytes * 8) / mdp->desc_len;
+    else
+        u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
+                       mdp->num_descs_bytes * 8) + mdp->num_descs_inc;
     num = (int)u;
     if (opts->verbose)
         fprintf(stderr, "    >>> mode page says it has %d descriptors\n",
@@ -550,6 +576,43 @@ print_mpage_extra_desc(void ** pc_arr, int rep_len,
     }
 }
 
+/* Print WP and DPOFUA bits for direct access devices (disks).
+ * Return 0 for success, else error value. */
+static int
+print_direct_access_info(int sg_fd, const struct sdparm_opt_coll * opts,
+                         int verb)
+{
+    int res, v;
+    unsigned char cur_mp[DEF_MODE_RESP_LEN];
+
+    memset(cur_mp, 0, sizeof(cur_mp));
+    res = ll_mode_sense(sg_fd, opts, ALL_MPAGES, 0, cur_mp, 8, 0, verb);
+    if (0 == res) {
+        v = cur_mp[opts->mode_6 ? 2 : 3];
+        printf("    Direct access device specific parameters: WP=%d  "
+               "DPOFUA=%d\n", !!(v & 0x80), !!(v & 0x10));
+    } else if (SG_LIB_CAT_NOT_READY == res) {
+        fprintf(stderr, "mode sense command failed, device not ready\n");
+        return res;
+    } else if (SG_LIB_CAT_INVALID_OP == res) {
+        if (opts->mode_6)
+            fprintf(stderr, "6 byte MODE SENSE cdb not supported, "
+                    "try again without '-6' option\n");
+        else
+            fprintf(stderr, "10 byte MODE SENSE cdb not supported, "
+                    "try again with '-6' option\n");
+        return res;
+    } else if (SG_LIB_CAT_UNIT_ATTENTION == res) {
+        fprintf(stderr, "mode sense command failed, unit attention\n");
+        return res;
+    } else if (SG_LIB_CAT_ABORTED_COMMAND == res) {
+        fprintf(stderr, "mode sense command failed, aborted command\n");
+        return res;
+    }
+    return 0;
+}
+
+/* Print one or more mode pages. Returns 0 if ok else IO error number.  */
 static int
 print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                  const struct sdparm_opt_coll * opts)
@@ -557,7 +620,7 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
     int res, len, verb, smask, single_pg, fetch_pg, rep_len, orig_pn, warned;
     int first_desc_off;
     const struct sdparm_mode_page_item * mpi;
-    const struct sdparm_mode_page_item * init_mpi;
+    const struct sdparm_mode_page_item * hmpi;
     const struct sdparm_mode_page_item * fdesc_mpi = NULL;
     const struct sdparm_mode_page_t * mpp = NULL;
     const struct sdparm_mode_descriptor_t * mdp;
@@ -572,32 +635,11 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
     buff[0] = '\0';
     verb = (opts->verbose > 0) ? opts->verbose - 1 : 0;
     if ((0 == pdt) && (opts->long_out > 0) && (0 == opts->quiet)) {
-        memset(cur_mp, 0, sizeof(cur_mp));
-        res = ll_mode_sense(sg_fd, opts, ALL_MPAGES, 0, cur_mp, 8, 0, verb);
-        if (0 == res) {
-            smask = cur_mp[opts->mode_6 ? 2 : 3];
-            printf("    Direct access device specific parameters: WP=%d  "
-                   "DPOFUA=%d\n", !!(smask & 0x80), !!(smask & 0x10));
-        } else if (SG_LIB_CAT_NOT_READY == res) {
-            fprintf(stderr, "mode sense command failed, device not ready\n");
+        if (0 != (res = print_direct_access_info(sg_fd, opts, verb)))
             return res;
-        } else if (SG_LIB_CAT_INVALID_OP == res) {
-            if (opts->mode_6)
-                fprintf(stderr, "6 byte MODE SENSE cdb not supported, "
-                        "try again without '-6' option\n");
-            else
-                fprintf(stderr, "10 byte MODE SENSE cdb not supported, "
-                        "try again with '-6' option\n");
-            return res;
-        } else if (SG_LIB_CAT_UNIT_ATTENTION == res) {
-            fprintf(stderr, "mode sense command failed, unit attention\n");
-            return res;
-        } else if (SG_LIB_CAT_ABORTED_COMMAND == res) {
-            fprintf(stderr, "mode sense command failed, aborted command\n");
-            return res;
-        }
     }
     orig_pn = pn;
+    /* choose a mode page item namespace (vendor, transport or generic) */
     if (opts->vendor >= 0) {
         const struct sdparm_vendor_pair * vpp;
 
@@ -610,8 +652,8 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
     if (NULL == mpi)
         return SG_LIB_CAT_OTHER;
 
-    init_mpi = mpi;
-    if (pn >= 0) {      /* given page so step to first field */
+    hmpi = mpi;
+    if (pn >= 0) {      /* step to first item of given page */
         single_pg = 1;
         fetch_pg = 1;
         for ( ; mpi->acron; ++mpi) {
@@ -623,7 +665,7 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
         }
         if (NULL == mpi->acron) {       /* page has no known fields */
             if (opts->hex)
-                mpi = init_mpi;    /* trick to enter main loop once */
+                mpi = hmpi;    /* trick to enter main loop once */
             else {
                 sdp_get_mpage_name(pn, spn, pdt, opts->transport,
                                    opts->vendor, 0, 0, buff, sizeof(buff));
@@ -631,13 +673,15 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                         "add '-H' to see page in hex\n", buff);
             }
         }
-    } else {
+    } else {    /* want all so check all items in given namespace */
         single_pg = 0;
         fetch_pg = 0;
-        mpi = init_mpi;
+        mpi = hmpi;
     }
     name = "";
     mdp = NULL;
+
+    /* loop over each mode page item in given namespace */
     for (smask = 0, len = 0, warned = 0 ; mpi->acron; ++mpi, fetch_pg = 0) {
         if (0 == fetch_pg) {
             if ((pdt >=0) && (mpi->pdt >= 0) &&
@@ -647,6 +691,12 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                    (MF_COMMON & mpi->flags)))
                 continue;
             if ((pn != mpi->page_num) || (spn != mpi->subpage_num)) {
+                if (fdesc_mpi) {
+                    hmpi = mpi - 1;
+                    if ((pn == hmpi->page_num) && (spn == hmpi->subpage_num))
+                        print_mpage_extra_desc(pc_arr, rep_len, mpp,
+                                               fdesc_mpi, hmpi, opts, smask);
+                }
                 if (single_pg)
                     break;
                 fetch_pg = 1;
@@ -655,6 +705,7 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
             }
         }
         if (fetch_pg) {
+            /* Only fetch mode page when needed (e.g. item page changed) */
             mpp = sdp_get_mpage_name(pn, spn, pdt, opts->transport,
                                      opts->vendor, opts->long_out, opts->hex,
                                      buff, sizeof(buff));
@@ -689,9 +740,10 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                         "command\n");
                 return res;
             }
-            mdp = (mpp && single_pg && (! opts->hex)) ? mpp->mp_desc : NULL;
+            /* check for mode page descriptors */
+            mdp = (mpp && (! opts->hex)) ? mpp->mp_desc : NULL;
             first_desc_off = mdp ? mdp->first_desc_off : 0;
-            if (first_desc_off > 3) {
+            if (first_desc_off > 1) {
                 for (res = 0, fdesc_mpi = mpi;
                      fdesc_mpi && (pn == fdesc_mpi->page_num) &&
                      (spn == fdesc_mpi->subpage_num); ++fdesc_mpi) {
@@ -708,9 +760,14 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                 uint64_t u;
 
                 if (fdesc_mpi && (smask & 1)) {
-                    u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
-                                           mdp->num_descs_bytes * 8) +
-                        mdp->num_descs_inc;
+                    if ((mdp->num_descs_inc < 0) && (mdp->desc_len > 0))
+                        u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
+                                               mdp->num_descs_bytes * 8) /
+                            mdp->desc_len;
+                    else
+                        u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
+                                               mdp->num_descs_bytes * 8) +
+                            mdp->num_descs_inc;
                     num = (int)u;
                 }
                 if (opts->long_out)
@@ -773,7 +830,7 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                         fprintf(stderr, "failed\n");
                 }
             }
-        }
+        }       /* end of fetch_pg */
         if (smask && (! opts->hex)) {
             if (mpi->start_byte >= len) {
                 if ((0 == opts->flexible) && (0 == opts->verbose))
@@ -792,11 +849,10 @@ print_mode_pages(int sg_fd, int pn, int spn, int pdt,
                 if (0 == opts->flexible)
                     continue;
             }
-            print_mp_entry("  ", smask, mpi, cur_mp, cha_mp, def_mp,
-                           sav_mp, opts->long_out, 0);
+            print_mp_arr_entry("  ", smask, mpi, pc_arr, opts->long_out, 0);
         }
     }
-    if (mpi && single_pg && fdesc_mpi) {
+    if (mpi && (NULL == mpi->acron) && fdesc_mpi) {
         --mpi;
         if ((pn == mpi->page_num) && (spn == mpi->subpage_num))
             print_mpage_extra_desc(pc_arr, rep_len, mpp, fdesc_mpi, mpi,
@@ -840,8 +896,12 @@ desc_adjust_start_byte(int desc_num, const struct sdparm_mode_page_t * mpp,
 
     mdp = mpp->mp_desc;
     if ((mdp->num_descs_off < rep_len) && (mdp->num_descs_off < 64)) {
-        u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
-            mdp->num_descs_bytes * 8) + mdp->num_descs_inc;
+        if ((mdp->num_descs_inc < 0) && (mdp->desc_len > 0))
+            u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
+                    mdp->num_descs_bytes * 8) / mdp->desc_len;
+        else
+            u = sdp_get_big_endian(cur_mp + mdp->num_descs_off, 7,
+                    mdp->num_descs_bytes * 8) + mdp->num_descs_inc;
         if ((uint64_t)desc_num < u) {
             if (mdp->desc_len > 0) {
                 mpi->start_byte += (mdp->desc_len * desc_num);
@@ -885,11 +945,13 @@ desc_adjust_start_byte(int desc_num, const struct sdparm_mode_page_t * mpp,
     return 0;
 }
 
+/* Print one or more mode page items (from '--get='). Returns 0 if ok. */
 static int
 print_mode_items(int sg_fd, const struct sdparm_mode_page_settings * mps,
                  int pdt, const struct sdparm_opt_coll * opts)
 {
     int k, res, verb, smask, pn, spn, warned, rep_len, len, desc_num, adapt;
+    int long_o;
     uint64_t u;
     int64_t val;
     const struct sdparm_mode_page_item * mpi;
@@ -906,15 +968,15 @@ print_mode_items(int sg_fd, const struct sdparm_mode_page_settings * mps,
 
     warned = 0;
     verb = (opts->verbose > 0) ? opts->verbose - 1 : 0;
+    long_o = opts->long_out;
     for (k = 0, pn = 0, spn = 0; k < mps->num_it_vals; ++k) {
         ivp = &mps->it_vals[k];
         val = ivp->val;
         desc_num = ivp->descriptor_num;
         mpi = &ivp->mpi;
         mpp = sdp_get_mpage_name(mpi->page_num, mpi->subpage_num, mpi->pdt,
-                                 opts->transport, opts->vendor,
-                                 opts->long_out, opts->hex, buff,
-                                 sizeof(buff));
+                                 opts->transport, opts->vendor, long_o,
+                                 opts->hex, buff, sizeof(buff));
         if (desc_num > 0) {
             if (check_desc_convert_mpi(desc_num, mpp, mpi, &ampi, b_tmp,
                                        sizeof(b_tmp))) {
@@ -1048,8 +1110,7 @@ print_mode_items(int sg_fd, const struct sdparm_mode_page_settings * mps,
                     printf("-    ");
                 printf("\n");
             } else
-                print_mp_entry("", smask, mpi, cur_mp, cha_mp,
-                               def_mp, sav_mp, opts->long_out, 0);
+                print_mp_arr_entry("", smask, mpi, pc_arr, long_o, 0);
         } else if (1 == val) {
             if (opts->hex) {
                 if (smask & 1) {
@@ -1059,8 +1120,7 @@ print_mode_items(int sg_fd, const struct sdparm_mode_page_settings * mps,
                     printf("-    ");
                 printf("\n");
             } else
-                print_mp_entry("", smask & 1, mpi, cur_mp, NULL,
-                               NULL, NULL, opts->long_out, 0);
+                print_mp_arr_entry("", smask & 1, mpi, pc_arr, long_o, 0);
         } else if (2 == val) {
             if (opts->hex) {
                 if (smask & 1) {
@@ -1070,8 +1130,7 @@ print_mode_items(int sg_fd, const struct sdparm_mode_page_settings * mps,
                     printf("-    ");
                 printf("\n");
             } else
-                print_mp_entry("", smask & 1, mpi, cur_mp, NULL,
-                               NULL, NULL, opts->long_out, 1);
+                print_mp_arr_entry("", smask & 1, mpi, pc_arr, long_o, 1);
         }
     }
     return 0;
@@ -1661,7 +1720,7 @@ open_and_simple_inquiry(const char * device_name, int rw, int * pdt,
     }
     res = sg_simple_inquiry(sg_fd, &sir, 0, verb);
     if (res) {
-#ifdef SDPARM_LINUX
+#ifdef SG_LIB_LINUX
         if (-1 == res) {
             int sg_sg_fd;
 
@@ -1674,7 +1733,7 @@ open_and_simple_inquiry(const char * device_name, int rw, int * pdt,
             if (sg_sg_fd < 0)
                 goto err_out;
         }
-#endif  /* SDPARM_LINUX */
+#endif  /* SG_LIB_LINUX */
         if (res) {
             fprintf(stderr, "SCSI INQUIRY command failed on %s\n",
                     device_name);
@@ -1729,10 +1788,11 @@ err_out:
     return -1;
 }
 
+/* Process mode page(s) operation. Returns 0 if successful */
 static int
-process_mode_page(int sg_fd, const struct sdparm_mode_page_settings * mps,
-                  int pn, int spn, int rw, int get,
-                  const struct sdparm_opt_coll * opts, int pdt)
+process_mode(int sg_fd, const struct sdparm_mode_page_settings * mps, int pn,
+             int spn, int rw, int get, const struct sdparm_opt_coll * opts,
+             int pdt)
 {
     int res;
     const struct sdparm_mode_page_t * mpp;
@@ -1740,6 +1800,7 @@ process_mode_page(int sg_fd, const struct sdparm_mode_page_settings * mps,
     if ((pn > 0x3e) || (spn > 0xfe)) {
         fprintf(stderr, "Allowable mode page numbers are 0 to 62\n");
         fprintf(stderr, "  Allowable mode subpage numbers are 0 to 254\n");
+        fprintf(stderr, "  For VPD pages add a '-i' command line option\n");
         return SG_LIB_SYNTAX_ERROR;
     }
     if ((pn > 0) && (pdt >= 0)) {
@@ -1753,6 +1814,9 @@ process_mode_page(int sg_fd, const struct sdparm_mode_page_settings * mps,
                     mpp->name);
             fprintf(stderr, "   peripheral device type 0x%x but device "
                     "pdt is 0x%x\n", mpp->pdt, pdt);
+            if (! opts->flexible)
+                fprintf(stderr, "   may need '--flexible' option to "
+                        "override\n");
         }
     }
     if (opts->defaults)
@@ -1802,7 +1866,7 @@ main(int argc, char * argv[])
     const char * ccp;
     const struct sdparm_command_t * scmdp = NULL;
     int ret = 0;
-#ifdef SDPARM_WIN32
+#ifdef SG_LIB_WIN32
     int do_wscan = 0;
 #endif
 
@@ -1815,11 +1879,11 @@ main(int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-#ifdef SDPARM_WIN32
-        c = getopt_long(argc, argv, "6aBc:C:dDefg:hHilM:np:qs:St:vVw",
+#ifdef SG_LIB_WIN32
+        c = getopt_long(argc, argv, "6aBc:C:dDefg:hHilM:np:qrs:St:vVw",
                         long_options, &option_index);
 #else
-        c = getopt_long(argc, argv, "6aBc:C:dDefg:hHilM:np:qs:St:vV",
+        c = getopt_long(argc, argv, "6aBc:C:dDefg:hHilM:np:qrs:St:vV",
                         long_options, &option_index);
 #endif
         if (c == -1)
@@ -1910,6 +1974,9 @@ main(int argc, char * argv[])
             } else
                 page_str = optarg;
             break;
+        case 'r':
+            ++opts.read_only;
+            break;
         case 's':
             set_str = optarg;
             rw = 1;
@@ -1946,7 +2013,7 @@ main(int argc, char * argv[])
         case 'V':
             fprintf(stderr, "version: %s\n", version_str);
             return 0;
-#ifdef SDPARM_WIN32
+#ifdef SG_LIB_WIN32
         case 'w':
             ++do_wscan;
             break;
@@ -1970,14 +2037,17 @@ main(int argc, char * argv[])
         }
     }
 
+    if (opts.read_only)
+        rw = 0;         // override any read-write settings
+
     if ((!!get_str + !!set_str + !!clear_str) > 1) {
         fprintf(stderr, "Can only give one of '--get=', '--set=' and "
                 "'--clear='\n");
         return SG_LIB_SYNTAX_ERROR;
     }
-#ifdef SDPARM_WIN32
+#ifdef SG_LIB_WIN32
     if (do_wscan)
-        return sg_do_wscan('\0', opts.verbose);
+        return sg_do_wscan('\0', do_wscan, opts.verbose);
 #endif
 
     if (page_str) {
@@ -2105,6 +2175,8 @@ main(int argc, char * argv[])
             sdp_enumerate_commands();
             return SG_LIB_SYNTAX_ERROR;
         }
+        if (opts.read_only)
+            rw = 0;         // override any read-write settings
     } else {
         /* assume mode pages */
         if (pn < 0) {
@@ -2290,8 +2362,8 @@ main(int argc, char * argv[])
             if (cmd_str && scmdp)   /* process command */
                 r = sdp_process_cmd(sg_fd, scmdp, cmd_arg, pdt, &opts);
             else                    /* mode page */
-                r = process_mode_page(sg_fd, &mp_settings, pn, spn, rw,
-                                      (NULL != get_str), &opts, pdt);
+                r = process_mode(sg_fd, &mp_settings, pn, spn, rw,
+                                 (NULL != get_str), &opts, pdt);
         }
 
         res = sg_cmds_close_device(sg_fd);
@@ -2306,8 +2378,13 @@ main(int argc, char * argv[])
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
 
-#ifdef SDPARM_LINUX
-/*     ============ */
+#ifdef SG_LIB_LINUX
+/* Following needed for lk 2.4 series; may be dropped in future.
+ * In the lk 2.4 series (and earlier) no pass-through was available on the
+ * often used /dev/sd* device nodes. So the code below attempts to map a
+ * given /dev/sd<n> device node to the corresponding /dev/sg<m> device
+ * node.
+ */
 
 typedef struct my_scsi_idlun
 {
@@ -2457,5 +2534,4 @@ static int map_if_lk24(int sg_fd, const char * device_name, int rw,
     return find_corresponding_sg_fd(sg_fd, device_name, rw, verbose);
 }
 
-#endif  /* SDPARM_LINUX */
-
+#endif  /* SG_LIB_LINUX */
