@@ -1,11 +1,10 @@
 #ifndef SDPARM_H
 #define SDPARM_H
 
-/* sdparm is a utility program for the Linux OS SCSI subsystem.
- *
- * This utility fetches various parameters associated with a given
- * SCSI disk (or a disk that uses, or translates the SCSI command
- * set). In some cases these parameters can be changed.
+/*
+ * sdparm is a utility program for getting and setting parameters on devices
+ * that use one of the SCSI command sets. In some cases commands can be sent
+ * to the device (e.g. eject removable media).
  */
 
 #include <stdint.h>
@@ -28,6 +27,7 @@
 #define CONTROL_MP 0xa
 #define POWER_OLD_MP 0xd
 /* #define CD_DEV_PARAMS 0xd */
+#define ADC_MP 0xe
 #define DATA_COMPR_MP 0xf
 #define DEV_CONF_MP 0x10
 #define XOR_MP 0x10
@@ -46,7 +46,6 @@
 #define ALL_MPAGES 0x3f
 
 /* Mode subpage numbers */
-#define MSP_CONTROL_EXT 1       /* unused */
 #define MSP_SPC_CE 1            /* control extension */
 #define MSP_SPI_MC 1
 #define MSP_SPI_STC 2
@@ -54,11 +53,18 @@
 #define MSP_SPI_RTC 4
 #define MSP_SAS_PCD 1
 #define MSP_SAS_SPC 2
-#define MSP_SAS2_PHY 3
+#define MSP_SAS_E_PHY 3
 #define MSP_BACK_CTL 1
 #define MSP_SAT_PATA 0xf1       /* SAT PATA Control */
+#define MSP_SAT_POWER 0xf1      /* SAT ATA Power condition */
 #define MSP_DEV_CONF_EXT 1      /* device conf extension (ssc) */
 #define MSP_EXT_DEV_CAP 0x41    /* extended device capabilities (smc) */
+#define MSP_ADC_TGT_DEV 0x1
+#define MSP_ADC_DT_DPP 0x2
+#define MSP_ADC_LU 0x3
+#define MSP_ADC_TD_SN 0x4
+#define MSP_SBC_THIN_PROV 0x2
+#define MSP_SSC_CDP 0xf0
 
 #define MODE_DATA_OVERHEAD 128
 #define EBUFF_SZ 256
@@ -78,6 +84,8 @@
 #define VPD_MODE_PG_POLICY 0x87
 #define VPD_SCSI_PORTS 0x88
 #define VPD_ATA_INFO 0x89
+#define VPD_POWER_CONDITION 0x8a
+#define VPD_DEVICE_CONSTITUENTS 0x8b
 #define VPD_PROTO_LU 0x90
 #define VPD_PROTO_PORT 0x91
 #define VPD_BLOCK_LIMITS 0xb0   /* SBC-3 */
@@ -87,6 +95,10 @@
 #define VPD_MAN_ASS_SN 0xb1     /* SSC-3, ADC-2 */
 #define VPD_SECURITY_TOKEN 0xb1 /* OSD */
 #define VPD_TA_SUPPORTED 0xb2   /* SSC-3 */
+#define VPD_THIN_PROVISIONING 0xb2   /* SBC-3 */
+#define VPD_REFERRALS 0xb3
+#define VPD_AUTOMATION_DEV_SN 0xb3   /* SSC-3 */
+#define VPD_DTDE_ADDRESS 0xb4   /* SSC-4 */
 
 #define VPD_ASSOC_LU 0
 #define VPD_ASSOC_TPORT 1
@@ -138,6 +150,7 @@ struct sdparm_opt_coll {
     int mode_6;
     int num_desc;
     int quiet;
+    int read_only;
     int save;
     int transport;
     int vendor;
@@ -145,15 +158,17 @@ struct sdparm_opt_coll {
 };
 
 struct sdparm_mode_descriptor_t {
-    int num_descs_off;
-    int num_descs_bytes;
-    int num_descs_inc;    /* number to add to num_descs derived from */
-                          /* first 2 entries */
+    int num_descs_off;    /* byte offset of start of num_descriptors */
+    int num_descs_bytes;  /* number of bytes in num_descriptors field */
+    int num_descs_inc;    /* number to add to num_descriptors */
+                          /* if negative then value in num_descriptors */
+                          /* is byte count, so divide by desc_len */
     int first_desc_off;
     int desc_len;         /* -1 for unknown otherwise fixed per desc */
     int desc_len_off;     /* if (-1 == desc_len) then this is offset */
     int desc_len_bytes;   /* ... after start of descriptor */
-    /* Hence: <desc_len> = *(d_len_off + base) + d_len_off + d_len_bytes */
+    /* Hence: <desc_len> = deref(base + d_len_off, d_len_bytes) + */
+    /*                     d_len_off + d_len_bytes */
     const char * name;
 };
 
@@ -327,9 +342,9 @@ extern int sdp_process_cmd(int sg_fd, const struct sdparm_command_t * scmdp,
  * Declarations for functions that are port dependant
  */
 
-#ifdef SDPARM_WIN32
+#ifdef SG_LIB_WIN32
 
-extern int sg_do_wscan(char letter, int verbose);
+extern int sg_do_wscan(char letter, int do_scan, int verbose);
 
 #endif
 
