@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005-2008 Douglas Gilbert.
+ * Copyright (c) 2005-2010 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -48,25 +48,53 @@ static struct sdparm_mode_descriptor_t ssc_mpa_desc = {
     3, 1, 1, 8, 2, -1, -1, "SSC medium partition"
 };
 
+/* SMC's transport geometry parameters mode page doesn't give the number
+   of following descriptors but rather parameter length (in bytes).
+   This is flagged by -1 in num_descs_inc (third) field */
+static struct sdparm_mode_descriptor_t smc_tg_desc = {
+    1, 1, -1, 2, 2, -1, -1, "SMC transport geometry"
+};
+
+/* SBC's thin provisioning mode page doesn't give the number
+   of following descriptors but rather parameter length (in bytes).
+   This is flagged by -1 in num_descs_inc (third) field */
+static struct sdparm_mode_descriptor_t sbc_tp_desc = {
+    2, 2, -1, 16, 8, -1, -1, "SBC thin provisioning"
+};
+
 /* Mode pages that aren't specific to any transport protocol or vendor.
    Note that all standard periperal device types are include.
    The pages are listed in acronym alphabetical order. */
 struct sdparm_mode_page_t sdparm_gen_mode_pg[] = {
-    {IEC_MP, MSP_BACK_CTL, 0, PDT_DISK, "bc", "Background control (SBC)",
+    {ADC_MP, MSP_ADC_DT_DPP, PDT_ADC, 0, "addp",
+        "DT device primary port (ADC)", NULL},
+    {ADC_MP, MSP_ADC_LU, PDT_ADC, 0, "adlu", "logical unit (ADC)", NULL},
+    {ADC_MP, MSP_ADC_TGT_DEV, PDT_ADC, 0, "adtd", "Targer device (ADC)",
+        NULL},
+    {ADC_MP, MSP_ADC_TD_SN, PDT_ADC, 0, "adts",
+        "Targer device serial number (ADC)", NULL},
+    {POWER_MP, MSP_SAT_POWER, -1, 0, "apo", "SAT ATA Power condition", NULL},
+    {IEC_MP, MSP_BACK_CTL, PDT_DISK, 0, "bc", "Background control (SBC)",
         NULL},
     {CACHING_MP, 0, PDT_DISK, 0, "ca", "Caching (SBC)", NULL},
     {MMCMS_MP, 0, PDT_MMC, 1, "cms", "CD/DVD (MM) capabilities and "
         "mechanical status (MMC)", NULL},        /* read only */
     {CONTROL_MP, 0, -1, 0, "co", "Control", NULL},
     {CONTROL_MP, MSP_SPC_CE, -1, 0, "coe", "Control extension", NULL},
+    {CONTROL_MP, MSP_SSC_CDP, -1, 0, "cdp", "Control data protection (SSC)",
+        NULL},
     {DATA_COMPR_MP, 0, PDT_TAPE, 0, "dac", "Data compression (SSC)", NULL},
     {DEV_CONF_MP, 0, PDT_TAPE, 0, "dc", "Device configuration (SSC)", NULL},
+    {DEV_CAP_MP, 0, PDT_MCHANGER, 0, "dca", "Device capabilities (SMC)",
+        NULL},
     {DEV_CONF_MP, MSP_DEV_CONF_EXT, PDT_TAPE, 0, "dce",
         "Device configuration extension (SSC)", NULL},
     {DISCONNECT_MP, 0, -1, 0, "dr",
         "Disconnect-reconnect (SPC + transports)", NULL},
     {ELE_ADDR_ASS_MP, 0, PDT_MCHANGER, 0, "eaa",
         "Element address assignment (SMC)", NULL},
+    {DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 0, "edc",
+        "Extended device capabilities (SMC)", NULL},
     {ES_MAN_MP, 0, PDT_SES, 0, "esm", "Enclosure services management (SES)",
         NULL},
     {FORMAT_MP, 0, PDT_DISK, 0, "fo", "Format (SBC)", NULL},
@@ -78,14 +106,18 @@ struct sdparm_mode_page_t sdparm_gen_mode_pg[] = {
     {CONTROL_MP, MSP_SAT_PATA, -1, 0, "pat", "SAT pATA control", NULL},
     {PROT_SPEC_LU_MP, 0, -1, 0, "pl", "Protocol specific logical unit", NULL},
     {POWER_MP, 0, -1, 0, "po", "Power condition", NULL},
-    {POWER_OLD_MP, 0, PDT_DISK, 0, "poo", "Power condition - old version", NULL},
-        /* POWER_OLD_MP for disks as it clashes with old MMC specs */
+    {POWER_OLD_MP, 0, PDT_DISK, 0, "poo", "Power condition - old version",
+        NULL}, /* POWER_OLD_MP for disks as it clashes with old MMC specs */
     {PROT_SPEC_PORT_MP, 0, -1, 0, "pp", "Protocol specific port", NULL},
     {RBC_DEV_PARAM_MP, 0, PDT_RBC, 0, "rbc", "RBC device parameters (RBC)",
         NULL},
     {RIGID_DISK_MP, 0, PDT_DISK, 0, "rd", "Rigid disk (SBC)", NULL},
     {RW_ERR_RECOVERY_MP, 0, -1, 0, "rw", "Read write error recovery", NULL},
         /* since in SBC, SSC and MMC treat RW_ERR_RECOVERY_MP as if in SPC */
+    {TRANS_GEO_PAR_MP, 0, PDT_MCHANGER, 0, "tgp", "Transport geometry "
+        "parameters (SMC)", &smc_tg_desc},
+    {IEC_MP, MSP_SBC_THIN_PROV, PDT_DISK, 0, "thp", "Thin provisioning "
+        "(SBC)", &sbc_tp_desc},
     {TIMEOUT_PROT_MP, 0, PDT_MMC, 0, "tp", "Timeout and protect (MMC)", NULL},
     {V_ERR_RECOVERY_MP, 0, PDT_DISK, 0, "ve", "Verify error recovery (SBC)",
         NULL},
@@ -155,10 +187,13 @@ static struct sdparm_mode_descriptor_t sas_pcd_desc = {   /* desc SAS-2 */
     7, 1, 0, 8, 48, -1, -1, "SAS phy"
 };
 
-static struct sdparm_mode_descriptor_t sas2_phy_desc = {  /* desc SAS-2 */
-    7, 1, 0, 8, -1, 2, 2, "SAS-2 phy"
+static struct sdparm_mode_descriptor_t sas_e_phy_desc = {  /* desc SAS-2 */
+    7, 1, 0, 8, -1, 2, 2, "Enhanced phy"
 };
 
+/* N.B. In SAS 2.1 the spec was split with the upper levels going into the
+ * SAS Protocol Layer (SPL) document. So now the SPL drafts are the
+ * relevant SAS references. */
 static struct sdparm_mode_page_t sdparm_sas_mode_pg[] = {    /* SAS-2 */
     {DISCONNECT_MP, 0, -1, 0, "dr", "Disconnect-reconnect (SAS)", NULL},
     {PROT_SPEC_LU_MP, 0, -1, 0, "pl", "Protocol specific logical unit (SAS)",
@@ -166,8 +201,8 @@ static struct sdparm_mode_page_t sdparm_sas_mode_pg[] = {    /* SAS-2 */
     {PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 0, "pcd",
         "Phy control and discover (SAS)", &sas_pcd_desc},
     {PROT_SPEC_PORT_MP, 0, -1, 0, "pp", "Protocol specific port (SAS)", NULL},
-    {PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 0, "s2p",
-        "SAS-2 Phy", &sas2_phy_desc},
+    {PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 0, "sep",
+        "Enhanced phy control (SAS)", &sas_e_phy_desc},
     {PROT_SPEC_PORT_MP, MSP_SAS_SPC, -1, 0, "spc",
         "Shared port control (SAS)", NULL},
     {0, 0, 0, 0, NULL, NULL, NULL},
@@ -178,9 +213,12 @@ struct sdparm_vpd_page_t sdparm_vpd_pg[] = {
     {VPD_ATA_INFO, 0, -1, "ai", "ATA information (SAT)"},
     {VPD_ASCII_OP_DEF, 0, -1, "aod",
      "ASCII implemented operating definition (obs)"},
+    {VPD_AUTOMATION_DEV_SN, 0, PDT_TAPE, "adsn", "Automation device serial "
+     "number (SSC)"},
     {VPD_BLOCK_LIMITS, 0, PDT_DISK, "bl", "Block limits (SBC)"},
     {VPD_BLOCK_DEV_CHARS, 0, PDT_DISK, "bdc", "Block device characteristics "
      "(SBC)"},
+    {VPD_DEVICE_CONSTITUENTS, 0, -1, "dc", "Device constituents"},
     {VPD_DEVICE_ID, 0, -1, "di", "Device identification"},
     {VPD_DEVICE_ID, VPD_DI_SEL_AS_IS, -1, "di_asis", "Like 'di' "
      "but designators ordered as found"},
@@ -191,6 +229,8 @@ struct sdparm_vpd_page_t sdparm_vpd_pg[] = {
     {VPD_DEVICE_ID, VPD_DI_SEL_TARGET, -1, "di_target", "Device "
      "identification, target device only"},
     {VPD_EXT_INQ, 0, -1, "ei", "Extended inquiry data"},
+    {VPD_DTDE_ADDRESS, 0, 1, "dtde",
+     "Data transfer device element address (SSC)"},
     {VPD_IMP_OP_DEF, 0, -1, "iod",
      "Implemented operating definition (obs)"},
     {VPD_MAN_ASS_SN, 0, PDT_TAPE, "mas",
@@ -200,9 +240,11 @@ struct sdparm_vpd_page_t sdparm_vpd_pg[] = {
     {VPD_MAN_NET_ADDR, 0, -1, "mna", "Management network addresses"},
     {VPD_MODE_PG_POLICY, 0, -1, "mpp", "Mode page policy"},
     {VPD_OSD_INFO, 0, PDT_OSD, "oi", "OSD information"},
+    {VPD_POWER_CONDITION, 0, -1, "pc", "Power condition"},
     {VPD_PROTO_LU, 0, -1, "pslu", "Protocol-specific logical unit "
      "information"},
     {VPD_PROTO_PORT, 0, -1, "pspo", "Protocol-specific port information"},
+    {VPD_REFERRALS, 0, PDT_DISK, "ref", "Referrals (SBC)"},
     {VPD_SA_DEV_CAP, 0, PDT_TAPE, "sad",
      "Sequential access device capabilities (SSC)"},
     {VPD_SOFTW_INF_ID, 0, -1, "sii", "Software interface identification"},
@@ -210,6 +252,7 @@ struct sdparm_vpd_page_t sdparm_vpd_pg[] = {
     {VPD_SCSI_PORTS, 0, -1, "sp", "SCSI ports"},
     {VPD_SUPPORTED_VPDS, 0, -1, "sv", "Supported VPD pages"},
     {VPD_TA_SUPPORTED, 0, PDT_TAPE, "tas", "TapeAlert supported flags (SSC)"},
+    {VPD_THIN_PROVISIONING, 0, PDT_DISK, "thpv", "Thin provisioning (SBC)"},
     {0, 0, 0, NULL, NULL},
 };
 
@@ -229,12 +272,14 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
         "Read continuous", "0: error recovery may cause delays\t"
         "1: transfer data without waiting for error recovery"},
     {"EER", RW_ERR_RECOVERY_MP, 0, -1, 2, 3, 1, 0,
-        "Enable early recovery", NULL},
+        "Enable early recovery",
+        "1: increase chance of mis-detection or mis-correction of error"},
     {"PER", RW_ERR_RECOVERY_MP, 0, -1, 2, 2, 1, MF_COMMON,
         "Post error", "0: do not post recovered errors\t"
-        "1: report recovered errors"},
+        "1: report recovered errors (via sense key: recovered error)"},
     {"DTE", RW_ERR_RECOVERY_MP, 0, -1, 2, 1, 1, 0,
-        "Data terminate on error", NULL},
+        "Data terminate on error",
+        "1: terminate data transfer when recovered error detected"},
     {"DCR", RW_ERR_RECOVERY_MP, 0, -1, 2, 0, 1, 0,
         "Disable correction", NULL},
     {"RRC", RW_ERR_RECOVERY_MP, 0, -1, 3, 7, 8, 0,
@@ -245,14 +290,16 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
         "Head offset count (obsolete)", NULL},
     {"DSOC", RW_ERR_RECOVERY_MP, 0, -1, 6, 7, 8, 0,
         "Data strobe offset count (obsolete)", NULL},
+    {"TPERE", RW_ERR_RECOVERY_MP, 0, 0, 7, 7, 1, 0, /* SBC */
+        "Thin provisioning error reporting enabled", NULL},
     {"EMCDR", RW_ERR_RECOVERY_MP, 0, 5, 7, 1, 2, 0, /* MMC */
         "Enhanced media certification and defect reporting", NULL},
     {"WRC", RW_ERR_RECOVERY_MP, 0, -1, 8, 7, 8, 0,
         "Write retry count", NULL},
-    {"ERTL", RW_ERR_RECOVERY_MP, 0, 5, 9, 7, 24, 0, /* MMC */
-        "Error reporting threshold length (blocks)", NULL},
+    {"ERWS", RW_ERR_RECOVERY_MP, 0, 5, 9, 7, 24, 0, /* MMC, was ERTL */
+        "Error reporting window size (blocks)", NULL},
     {"RTL", RW_ERR_RECOVERY_MP, 0, 0, 10, 7, 16, 0, /* SBC */
-        "Recovery time limit (ms)", NULL},
+        "Recovery time limit (ms)", "0: default, -1: 65.5 seconds"},
 
     /* Disconnect-reconnect mode page [0x2]: spc-4 + */
     /* See transport sections for more detailed information */
@@ -467,6 +514,8 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
         "1: lu maintains separate task sets for each I_T nexus"},
     {"TMF_ONLY", CONTROL_MP, 0, -1, 2, 4, 1, 0,
         "Task management functions only", NULL},
+    {"DPICZ", CONTROL_MP, 0, -1, 2, 3, 1, 0,
+        "Disable protection information check if protect field zero", NULL},
     {"D_SENSE", CONTROL_MP, 0, -1, 2, 2, 1, 0,
         "Descriptor format sense data", NULL},
     {"GLTSD", CONTROL_MP, 0, -1, 2, 1, 1, 0,
@@ -476,6 +525,8 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
     {"QAM", CONTROL_MP, 0, -1, 3, 7, 4, 0,
         "Queue algorithm modifier",
         "0: restricted re-ordering; 1: unrestricted"},
+    {"NUAR", CONTROL_MP, 0, -1, 3, 3, 1, 0,
+        "No unit attention on release", NULL},
     {"QERR", CONTROL_MP, 0, -1, 3, 2, 2, 0,
         "Queue error management",
         "0: only affected task gets CC; 1: affected tasks aborted\t"
@@ -518,6 +569,19 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
         "Initial command priority", "0: none or vendor\t"
         "1: highest\t15: lowest"},
 
+    /* Control data protection mode subpage [0xa,0xf0] ssc4 */
+    {"LBPM", CONTROL_MP, MSP_SSC_CDP, PDT_TAPE, 4, 7, 8, 0,
+        "Logical block protection method", "0: none\t"
+        "1: Reed-Solomon CRC"},
+    {"LBPIL", CONTROL_MP, MSP_SSC_CDP, PDT_TAPE, 5, 5, 6, 0,
+        "Logical block protection information length", NULL},
+    {"LBP_W", CONTROL_MP, MSP_SSC_CDP, PDT_TAPE, 6, 7, 1, 0,
+        "Logical block protection during write", NULL},
+    {"LBP_R", CONTROL_MP, MSP_SSC_CDP, PDT_TAPE, 6, 6, 1, 0,
+        "Logical block protection during read", NULL},
+    {"RBDP", CONTROL_MP, MSP_SSC_CDP, PDT_TAPE, 6, 5, 1, 0,
+        "Recover buffered data protected", NULL},
+
     /* SAT: pATA control mode subpage [0xa,0xf1] sat-r09 */
     /* treat as spc since could be disk or ATAPI */
     {"MWD2", CONTROL_MP, MSP_SAT_PATA, -1, 4, 6, 1, 0,
@@ -699,15 +763,37 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
         "0: fcp; 1: spi; 4: srp; 5: iscsi; 6: sas; 7: adt; 8: ata/atapi\t"
         "[try adding '-t <transport>' to get more fields]"},
 
-    /* Power condition mode page [0x1a] spc3 */
-    {"IDLE", POWER_MP, 0, -1, 3, 1, 1, 0,
-        "Idle timer active", NULL},
-    {"STANDBY", POWER_MP, 0, -1, 3, 0, 1, 0,
-        "Standby timer active", NULL},
-    {"ICT", POWER_MP, 0, -1, 4, 7, 32, 0,
-        "Idle condition timer (100 ms)", NULL},
-    {"SCT", POWER_MP, 0, -1, 8, 7, 32, 0,
-        "Standby condition timer (100 ms)", NULL},
+    /* Power condition mode page [0x1a] spc3 (expanded in spc4r18) */
+    {"STANDBY_Y", POWER_MP, 0, -1, 2, 0, 1, 0,
+        "Standby_y timer enabled", NULL},
+    {"IDLE_C", POWER_MP, 0, -1, 3, 3, 1, 0,
+        "Idle_c timer enabled", NULL},
+    {"IDLE_B", POWER_MP, 0, -1, 3, 2, 1, 0,
+        "Idle_b timer active", NULL},
+    {"IDLE", POWER_MP, 0, -1, 3, 1, 1, 0,   /* IDLE_A in future ? */
+        "Idle timer enabled",
+        "named IDLE prior to spc4r18, thence IDLE_A"},
+    {"STANDBY", POWER_MP, 0, -1, 3, 0, 1, 0, /* STANDBY_A in future ? */
+        "Standby timer active",
+        "named STANDBY prior to spc4r18, thence STANDBY_Z"},
+    {"ICT", POWER_MP, 0, -1, 4, 7, 32, 0,    /* IACT in future ? */
+        "Idle condition timer (100 ms)",
+        "named IDLE prior to spc4r18, thence IDLE_A"},
+    {"SCT", POWER_MP, 0, -1, 8, 7, 32, 0,    /* SZCT in future ? */
+        "Standby condition timer (100 ms)",
+        "named STANDBY prior to spc4r18, thence STANDBY_Z"},
+    {"IBCT", POWER_MP, 0, -1, 12, 7, 32, 0,
+        "Idle_b condition timer (100 ms)", NULL},
+    {"ICCT", POWER_MP, 0, -1, 16, 7, 32, 0,
+        "Idle_c condition timer (100 ms)", NULL},
+    {"SYCT", POWER_MP, 0, -1, 20, 7, 32, 0,
+        "Standby_y condition timer (100 ms)", NULL},
+
+    /* SAT ATA Power condition mode page [0x1a,0xf1] sat2 */
+    {"APMP", POWER_MP, MSP_SAT_POWER, -1, 5, 0, 1, 0,
+        "Advanced Power Management (APM) enabled/change", NULL},
+    {"APM", POWER_MP, MSP_SAT_POWER, -1, 6, 7, 8, 0,
+        "Advanced Power Management (APM) value", NULL},
 
     /* Informational exception control mode page [0x1c] spc3 */
     {"PERF", IEC_MP, 0, -1, 2, 7, 1, 0,
@@ -753,6 +839,18 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
     {"MAX_SUSP", IEC_MP, MSP_BACK_CTL, PDT_DISK, 12, 7, 16, 0,
         "Maximum time to suspend background scan (ms)", NULL},
 
+    /* Thin provisioning mode subpage [0x1c,0x2] sbc3 */
+    {"TH_EN", IEC_MP, MSP_SBC_THIN_PROV, PDT_DISK, 16, 7, 1, 0,
+        "Threshold enabled", NULL},
+    {"TH_TYPE", IEC_MP, MSP_SBC_THIN_PROV, PDT_DISK, 16, 5, 3, 0,
+        "Threshold type", NULL},
+    {"TH_ARM", IEC_MP, MSP_SBC_THIN_PROV, PDT_DISK, 16, 2, 3, 0,
+        "Threshold arming", NULL},
+    {"TH_RES", IEC_MP, MSP_SBC_THIN_PROV, PDT_DISK, 17, 7, 8, 0,
+        "Threshold resource", NULL},
+    {"TH_COUNT", IEC_MP, MSP_SBC_THIN_PROV, PDT_DISK, 20, 7, 32, 0,
+        "Threshold count", NULL},
+
     /* Medium configuration mode page [0x1d] ssc3 */
     {"WORMM", MED_CONF_MP, 0, PDT_TAPE, 2, 0, 1, 0,
         "Worm mode", NULL},
@@ -798,6 +896,146 @@ struct sdparm_mode_page_item sdparm_mitem_arr[] = {
     {"NDTE", ELE_ADDR_ASS_MP, 0, PDT_MCHANGER, 16, 7, 16, 0,
         "Number of data transfer elements", NULL},
 
+    /* Transport geometry parameters mode page [0x1e] smc2 */
+    {"ROTAT", TRANS_GEO_PAR_MP, 0, PDT_MCHANGER, 2, 0, 1, 0,
+        "Rotation for double sided media handling", NULL},
+    {"MNTES", TRANS_GEO_PAR_MP, 0, PDT_MCHANGER, 3, 7, 8, 0,
+        "Member number in transport element set", NULL},
+
+    /* Device capabilities mode page [0x1f] smc3 */
+    {"STORDT", DEV_CAP_MP, 0, PDT_MCHANGER, 2, 3, 1, 0,
+        "Storage for data transfer element", NULL},
+    {"STORIE", DEV_CAP_MP, 0, PDT_MCHANGER, 2, 2, 1, 0,
+        "Storage for import/export element", NULL},
+    {"STORST", DEV_CAP_MP, 0, PDT_MCHANGER, 2, 1, 1, 0,
+        "Storage for storage element", NULL},
+    {"STORMT", DEV_CAP_MP, 0, PDT_MCHANGER, 2, 0, 1, 0,
+        "Storage for medium transport element", NULL},
+    {"ACE", DEV_CAP_MP, 0, PDT_MCHANGER, 3, 2, 1, 0,
+        "Auto clean enabled", NULL},
+    {"VTRP", DEV_CAP_MP, 0, PDT_MCHANGER, 3, 1, 1, 0,
+        "Volume tag reader present", NULL},
+    {"S2C", DEV_CAP_MP, 0, PDT_MCHANGER, 3, 0, 1, 0,
+        "SMC-2 capabilities supported", NULL},
+    {"MT_RA", DEV_CAP_MP, 0, PDT_MCHANGER, 4, 7, 2, 0,
+        "Medium transport elements support Read Attribute", NULL},
+    {"MT2DT", DEV_CAP_MP, 0, PDT_MCHANGER, 4, 3, 1, 0,
+        "Medium transport -> data transfer; Move Medium", NULL},
+    {"MT2IE", DEV_CAP_MP, 0, PDT_MCHANGER, 4, 2, 1, 0,
+        "Medium transport -> import/export; Move Medium", NULL},
+    {"MT2ST", DEV_CAP_MP, 0, PDT_MCHANGER, 4, 1, 1, 0,
+        "Medium transport -> storage; Move Medium", NULL},
+    {"MT2MT", DEV_CAP_MP, 0, PDT_MCHANGER, 4, 0, 1, 0,
+        "Medium transport -> medium transport; Move Medium", NULL},
+    {"ST_RA", DEV_CAP_MP, 0, PDT_MCHANGER, 5, 7, 2, 0,
+        "Storage elements support Read Attribute", NULL},
+    {"ST2DT", DEV_CAP_MP, 0, PDT_MCHANGER, 5, 3, 1, 0,
+        "Storage -> data transfer; Move Medium", NULL},
+    {"ST2IE", DEV_CAP_MP, 0, PDT_MCHANGER, 5, 2, 1, 0,
+        "Storage -> import/export; Move Medium", NULL},
+    {"ST2ST", DEV_CAP_MP, 0, PDT_MCHANGER, 5, 1, 1, 0,
+        "Storage -> storage; Move Medium", NULL},
+    {"ST2MT", DEV_CAP_MP, 0, PDT_MCHANGER, 5, 0, 1, 0,
+        "Storage -> medium transport; Move Medium", NULL},
+    {"IE_RA", DEV_CAP_MP, 0, PDT_MCHANGER, 6, 7, 2, 0,
+        "Import/export elements support Read Attribute", NULL},
+    {"IE2DT", DEV_CAP_MP, 0, PDT_MCHANGER, 6, 3, 1, 0,
+        "Import/export -> data transfer; Move Medium", NULL},
+    {"IE2IE", DEV_CAP_MP, 0, PDT_MCHANGER, 6, 2, 1, 0,
+        "Import/export -> import/export; Move Medium", NULL},
+    {"IE2ST", DEV_CAP_MP, 0, PDT_MCHANGER, 6, 1, 1, 0,
+        "Import/export -> storage; Move Medium", NULL},
+    {"IE2MT", DEV_CAP_MP, 0, PDT_MCHANGER, 6, 0, 1, 0,
+        "Import/export -> medium transport; Move Medium", NULL},
+    {"DT_RA", DEV_CAP_MP, 0, PDT_MCHANGER, 7, 7, 2, 0,
+        "Data transfer elements support Read Attribute", NULL},
+    {"DT2DT", DEV_CAP_MP, 0, PDT_MCHANGER, 7, 3, 1, 0,
+        "Data transfer -> data transfer; Move Medium", NULL},
+    {"DT2IE", DEV_CAP_MP, 0, PDT_MCHANGER, 7, 2, 1, 0,
+        "Data transfer -> import/export; Move Medium", NULL},
+    {"DT2ST", DEV_CAP_MP, 0, PDT_MCHANGER, 7, 1, 1, 0,
+        "Data transfer -> storage; Move Medium", NULL},
+    {"DT2MT", DEV_CAP_MP, 0, PDT_MCHANGER, 7, 0, 1, 0,
+        "Data transfer -> medium transport; Move Medium", NULL},
+    {"MT_WA", DEV_CAP_MP, 0, PDT_MCHANGER, 12, 7, 2, 0,
+        "Medium transport elements support Write Attribute", NULL},
+    {"MTEDT", DEV_CAP_MP, 0, PDT_MCHANGER, 12, 3, 1, 0,
+        "Medium transport -> data transfer; Exchange Medium", NULL},
+    {"MTEIE", DEV_CAP_MP, 0, PDT_MCHANGER, 12, 2, 1, 0,
+        "Medium transport -> import/export; Exchange Medium", NULL},
+    {"MTEST", DEV_CAP_MP, 0, PDT_MCHANGER, 12, 1, 1, 0,
+        "Medium transport -> storage; Exchange Medium", NULL},
+    {"MTEMT", DEV_CAP_MP, 0, PDT_MCHANGER, 12, 0, 1, 0,
+        "Medium transport -> medium transport; Exchange Medium", NULL},
+    {"ST_WA", DEV_CAP_MP, 0, PDT_MCHANGER, 13, 7, 2, 0,
+        "Storage elements support Write Attribute", NULL},
+    {"STEDT", DEV_CAP_MP, 0, PDT_MCHANGER, 13, 3, 1, 0,
+        "Storage -> data transfer; Exchange Medium", NULL},
+    {"STEIE", DEV_CAP_MP, 0, PDT_MCHANGER, 13, 2, 1, 0,
+        "Storage -> import/export; Exchange Medium", NULL},
+    {"STEST", DEV_CAP_MP, 0, PDT_MCHANGER, 13, 1, 1, 0,
+        "Storage -> storage; Exchange Medium", NULL},
+    {"STEMT", DEV_CAP_MP, 0, PDT_MCHANGER, 13, 0, 1, 0,
+        "Storage -> medium transport; Exchange Medium", NULL},
+    {"IE_WA", DEV_CAP_MP, 0, PDT_MCHANGER, 14, 7, 2, 0,
+        "Import/export elements support Write Attribute", NULL},
+    {"IEEDT", DEV_CAP_MP, 0, PDT_MCHANGER, 14, 3, 1, 0,
+        "Import/export -> data transfer; Exchange Medium", NULL},
+    {"IEEIE", DEV_CAP_MP, 0, PDT_MCHANGER, 14, 2, 1, 0,
+        "Import/export -> import/export; Exchange Medium", NULL},
+    {"IEEST", DEV_CAP_MP, 0, PDT_MCHANGER, 14, 1, 1, 0,
+        "Import/export -> storage; Exchange Medium", NULL},
+    {"IEEMT", DEV_CAP_MP, 0, PDT_MCHANGER, 14, 0, 1, 0,
+        "Import/export -> medium transport; Exchange Medium", NULL},
+    {"DT_WA", DEV_CAP_MP, 0, PDT_MCHANGER, 15, 7, 2, 0,
+        "Data transfer elements support Write Attribute", NULL},
+    {"DTEDT", DEV_CAP_MP, 0, PDT_MCHANGER, 15, 3, 1, 0,
+        "Data transfer -> data transfer; Exchange Medium", NULL},
+    {"DTEIE", DEV_CAP_MP, 0, PDT_MCHANGER, 15, 2, 1, 0,
+        "Data transfer -> import/export; Exchange Medium", NULL},
+    {"DTEST", DEV_CAP_MP, 0, PDT_MCHANGER, 15, 1, 1, 0,
+        "Data transfer -> storage; Exchange Medium", NULL},
+    {"DTEMT", DEV_CAP_MP, 0, PDT_MCHANGER, 15, 0, 1, 0,
+        "Data transfer -> medium transport; Exchange Medium", NULL},
+
+    /* Extended device capabilities mode page [0x1f,0x41] smc3 */
+    {"MVPRV", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 4, 5, 1, 0,
+        "Move prevented to import/export element", NULL},
+    {"MVCL", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 4, 4, 1, 0,
+        "Move closes import/export element", NULL},
+    {"MVOP", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 4, 3, 1, 0,
+        "Move opens import/export element", NULL},
+    {"USRCL", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 4, 2, 1, 0,
+        "User control import/export element close", NULL},
+    {"USROP", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 4, 1, 1, 0,
+        "User control import/export element open", NULL},
+    {"IEST", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 4, 0, 1, 0,
+        "Import/export element state", NULL},
+    {"DTETA", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 5, 4, 1, 0,
+        "Data transfer element empty on door access", NULL},
+    {"RSSEA", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 5, 3, 1, 0,
+        "Return to source storage element address", NULL},
+    {"MVTRY", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 5, 2, 1, 0,
+        "Move tray", NULL},
+    {"IEMGZ", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 5, 1, 1, 0,
+        "Import/export magazine", NULL},
+    {"SMGZ", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 5, 0, 1, 0,
+        "Storage magazine", NULL},
+    {"TREXC", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 6, 2, 1, 0,
+        "True exchange capable", NULL},
+    {"LCKIE", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 6, 1, 1, 0,
+        "Lock import/export element", NULL},
+    {"LCKD", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 6, 0, 1, 0,
+        "Lock door", NULL},
+    {"SPMER", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 7, 2, 1, 0,
+        "Source pre-move eject required", NULL},
+    {"DPMER", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 7, 1, 1, 0,
+        "Destination pre-move eject required", NULL},
+    {"PEPOS", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 7, 0, 1, 0,
+        "Pre-eject position", NULL},
+    {"UCST", DEV_CAP_MP, MSP_EXT_DEV_CAP, PDT_MCHANGER, 8, 0, 1, 0,
+        "Unassigned cleaning storage", NULL},
+
     /* CD/DVD (MM) capabilities and mechanical status mode page */
     /* [0x2a] obsolete in mmc4 and mmc5, last valid in mmc3 */
     /* MRSS field was already obsolete in mmc3 */
@@ -1185,7 +1423,7 @@ static struct sdparm_mode_page_item sdparm_mitem_sas_arr[] = {
     {"NPLR", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 13, 3, 4, 0,
         "Negotiated logical link rate",         /* sas2r07 */
         "0: unknown; 1: disabled; 2: phy reset problem; 3: spinup hold\t"
-        "4: port selector; 8: 1.5 Gbps; 9: 3 Gbps; 0xa: 6 Gbps"},
+        "4: port selector; 8: 1.5 Gbps; 9: 3 Gbps; 10: 6 Gbps"},
     {"ASIP", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 14, 3, 1, 0,
         "Attached SSP initiator port", NULL},
     {"ATIP", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 14, 2, 1, 0,
@@ -1198,24 +1436,24 @@ static struct sdparm_mode_page_item sdparm_mitem_sas_arr[] = {
         "Attached STP target port", NULL},
     {"AMTP", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 15, 1, 1, 0,
         "Attached SMP target port", NULL},
-    {"SASA", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 16, 7, 64, MF_HEX,
-        "SAS address", NULL},
-    {"ASASA", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 24, 7, 64, MF_HEX,
-        "Attached SAS address", NULL},
+    {"SASA", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 16, 7, 64,
+         MF_HEX | MF_COMMON, "SAS address", NULL},
+    {"ASASA", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 24, 7, 64,
+         MF_HEX | MF_COMMON, "Attached SAS address", NULL},
     {"APHID", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 32, 7, 8, 0,
         "Attached phy identifier", NULL},
     {"PMILR", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 40, 7, 4, 0,
         "Programmed minimum link rate",
-        "0: not programmed; 8: 1.5 Gbps; 9: 3 Gbps; 0xa: 6 Gbps"},
+        "0: not programmed; 8: 1.5 Gbps; 9: 3 Gbps; 10: 6 Gbps"},
     {"HMILR", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 40, 3, 4, 0,
         "Hardware minimum link rate",
-        "8: 1.5 Gbps; 9: 3 Gbps; 0xa: 6 Gbps"},
+        "8: 1.5 Gbps; 9: 3 Gbps; 10: 6 Gbps"},
     {"PMALR", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 41, 7, 4, 0,
         "Programmed maximum link rate",
-        "0: not programmed; 8: 1.5 Gbps; 9: 3 Gbps; 0xa: 6 Gbps"},
+        "0: not programmed; 8: 1.5 Gbps; 9: 3 Gbps; 10: 6 Gbps"},
     {"HMALR", PROT_SPEC_PORT_MP, MSP_SAS_PCD, -1, 41, 3, 4, 0,
         "Hardware maximum link rate",
-        "8: 1.5 Gbps; 9: 3 Gbps; 0xa: 6 Gbps"},
+        "8: 1.5 Gbps; 9: 3 Gbps; 10: 6 Gbps"},
 
     /* shared port control mode page [0x19,0x2] sas2 */
     {"PPID_2", PROT_SPEC_PORT_MP, MSP_SAS_SPC, -1, 5, 3, 4, 0,
@@ -1224,30 +1462,34 @@ static struct sdparm_mode_page_item sdparm_mitem_sas_arr[] = {
     {"PLT", PROT_SPEC_PORT_MP, MSP_SAS_SPC, -1, 6, 7, 16, 0,
         "Power loss timeout(ms)", NULL},
 
-    /* SAS-2 phy mode page [0x19,0x3] sas2 */
-    {"PPID_3", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 5, 3, 4, 0,
+    /* SAS-2 Enhanced phy mode page [0x19,0x3] sas2, spl */
+    {"PPID_3", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 5, 3, 4, 0,
         "Port's (transport) protocol identifier",
         "0: fcp; 1: spi; 4: srp; 5: iscsi; 6: sas; 7: adt; 8: ata/atapi"},
-    {"GENC_1", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 6, 7, 8, 0,
+    {"GENC_1", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 6, 7, 8, 0,
         "Generation code", "0: unknown, 1..255: valid"},
-    {"NOP_1", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 7, 7, 8, 0,
+    {"NOP_1", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 7, 7, 8, 0,
         "Number of phys", "one descriptor per phy"},
     /* */
-    {"PHID_1", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 9, 7, 8, 0,
+    {"PHID_1", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 9, 7, 8, 0,
         "Phy identifier", NULL},
-    {"PPCAP", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 12, 7, 32, MF_HEX,
+    {"PPCAP", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 12, 7, 32, MF_HEX,
         "Programmed phy capabilities", NULL},
-    {"CPCAP", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 16, 7, 32, MF_HEX,
+    {"CPCAP", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 16, 7, 32, MF_HEX,
         "Current phy capabilities", NULL},
-    {"APCAP", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 20, 7, 32, MF_HEX,
+    {"APCAP", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 20, 7, 32, MF_HEX,
         "Attached phy capabilities", NULL},
-    {"N_SSC", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 26, 4, 1, 0,
+    {"N_SSC", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 26, 4, 1, 0,
         "Negotiated spread spectrum clocking", NULL},
-    {"N_PLR", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 26, 3, 4, 0,
+    {"N_PLR", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 26, 3, 4, 0,
         "Negotiated physical link rate",
         "6: resetting; 7: attached unsupported\t"
-        "8: 1.5 Gbps; 9: 3 Gbps; 0xa: 6 Gbps"},
-    {"HMS", PROT_SPEC_PORT_MP, MSP_SAS2_PHY, -1, 27, 0, 1, 0,
+        "8: 1.5 Gbps; 9: 3 Gbps; 10: 6 Gbps"},
+    {"EN_SL", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 27, 2, 1, 0,
+        "Enable slumber phy power condition", NULL},
+    {"EN_PA", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 27, 1, 1, 0,
+        "Enable partial phy power condition", NULL},
+    {"HMS", PROT_SPEC_PORT_MP, MSP_SAS_E_PHY, -1, 27, 0, 1, 0,
         "Hardware muxing supported", NULL},
 
     {NULL, 0, 0, 0, 0, 0, 0, 0, NULL, NULL},
@@ -1357,7 +1599,8 @@ const char * sdparm_network_service_type_arr[] =
     "status",
     "logging",
     "code download",
-    "reserved[0x6]", "reserved[0x7]", "reserved[0x8]", "reserved[0x9]",
+    "administrative configuration service",
+    "reserved[0x7]", "reserved[0x8]", "reserved[0x9]",
     "reserved[0xa]", "reserved[0xb]", "reserved[0xc]", "reserved[0xd]",
     "reserved[0xe]", "reserved[0xf]", "reserved[0x10]", "reserved[0x11]",
     "reserved[0x12]", "reserved[0x13]", "reserved[0x14]", "reserved[0x15]",
